#lang sicp

(#%require "environment.rkt"
           "common.rkt"
           "syntax.rkt")

(#%provide eval-if
           eval-sequence
           eval-assignment
           eval-definition
           eval-unbind
           eval-and
           eval-or
           analyze-self-evaluating
           analyze-quoted
           analyze-variable
           analyze-if
           analyze-sequence
           analyze-assignment
           analyze-definition
           analyze-application
           analyze-lambda)

;; Evaulation of special forms

(define (eval-if exp env eproc)
  (if (true? (eproc (if-predicate exp) env))
      (eproc (if-consequent exp) env)
      (eproc (if-alternative exp) env)))

(define (eval-sequence exps env eproc)
  (cond ((last-exp? exps) (eproc (first-exp exps) env))
        (else (eproc (first-exp exps) env)
              (eval-sequence (rest-exps exps) env))))

(define (eval-assignment exp env eproc)
  (set-variable-value! (assignment-variable exp)
                       (eproc (assignment-value exp) env)
                       env)
  'ok)

(define (eval-definition exp env eproc)
  (define-variable! (definition-variable exp)
    (eproc (definition-value exp) env)
    env)
  'ok)

(define (eval-unbind exp env eproc)
  (make-unbound-variable! (unbound-variable exp) env))

(define (eval-and conjuncts env eproc)
  (cond ((null? conjuncts)
         'true)
        ((last-conjunct? conjuncts)
         (eproc (first-conjunct conjuncts) env))
        ((false? (eproc (first-conjunct conjuncts) env))
         'false)
        (else
         (eval-and (make-and (rest-conjuncts conjuncts)) env))))

(define (eval-or disjuncts env eproc)
  (if (null? disjuncts)
      'false
      (let ((val (eproc (first-disjunct disjuncts) env)))
        (if val
            val
            (eval-or (make-or (rest-disjuncts disjuncts)) env)))))


;; Analysis of special forms

(define (analyze-self-evaluating exp analyze-proc)
  (lambda (env) exp))

(define (analyze-quoted exp analyze-proc)
  (let ((qval (text-of-quotation exp)))
    (lambda (env) qval)))

(define (analyze-variable exp analyze-proc)
  (lambda (env) (lookup-variable-value exp env)))

(define (analyze-assignment exp analyze-proc)
  (let ((var (assignment-variable exp))
        (vproc (analyze-proc (assignment-value exp)
                             analyze-proc)))
    (lambda (env)
      (set-variable-value! var (vproc env) env)
      'ok)))

(define (analyze-definition exp analyze-proc)
  (let ((var (definition-variable exp))
        (vproc (analyze-proc (definition-value exp)
                             analyze-proc)))
    (lambda (env)
      (define-variable! var (vproc env) env)
      'ok)))

(define (analyze-if exp analyze-proc)
  (let ((pproc (analyze-proc (if-predicate exp) analyze-proc))
        (cproc (analyze-proc (if-consequent exp) analyze-proc))
        (aproc (analyze-proc (if-alternative exp) analyze-proc)))
    (lambda (env)
      (if (true? (pproc env))
          (cproc env)
          (aproc env)))))

(define (analyze-lambda exp analyze-proc)
  (let ((vars (lambda-parameters exp))
        (bproc (analyze-sequence
                (scan-out-defines
                 (lambda-body exp))
                analyze-proc)))
    (lambda (env) (make-procedure vars bproc env))))

(define (analyze-sequence exps analyze-proc)
  (define (sequentially proc1 proc2)
    (lambda (env) (proc1 env) (proc2 env)))
  (define (loop first-proc rest-procs)
    (if (null? rest-procs)
        first-proc
        (loop (sequentially first-proc (car rest-procs))
              (cdr rest-procs))))
  (let ((procs (map (lambda (e) (analyze-proc e analyze-proc))
                    exps)))
    (if (null? procs)
        (error "Empty sequence -- ANALYZE"))
    (loop (car procs) (cdr procs))))

(define (analyze-application exp analyze-proc)
  (let ((fproc (analyze-proc (operator exp) analyze-proc))
        (aprocs (map (lambda (e) (analyze-proc e analyze-proc))
                     (operands exp))))
    (lambda (env)
      (execute-application (fproc env)
                           (map (lambda (aproc) (aproc env))
                                aprocs)))))

(define (execute-application proc args)
  (cond ((primitive-procedure? proc)
         (apply-primitive-procedure proc args))
        ((compound-procedure? proc)
         ((procedure-body proc)
          (extend-environment (procedure-parameters proc)
                              args
                              (procedure-environment proc))))
        (else
         (error
          "Unknown procedure type -- EXECUTE-APPLICATION"
          proc))))
